/*
 * QR Code generator library (Rust)
 *
 * Copyright (c) Kanashius (https://www.kanashius.com)
 */
extern crate qrcodegen;

use qrcodegen::{DataTooLong, QrCode, QrCodeEcc};

#[repr(C)]
pub struct QRCode {
    pub error: u8,
    pub width: u32,
    pub height: u32,
    pub data: *mut bool,
    pub needed_bits: u32,
    pub possible_bits: u32,
}

#[no_mangle]
pub extern "C" fn EncodeText(input: *const u8, size: u32, ecl: u8, result: *mut QRCode) -> u8 {
    Clean(result);
    let input_option;
    unsafe{
        input_option = std::str::from_utf8(std::slice::from_raw_parts(input, size as usize));
    }
    let ecl = ecl_from_int(ecl);
    match input_option {
        Ok(text) => { wrap_qr_code(QrCode::encode_text(text, ecl), result) }
        Err(_) => {
            let mut data: Vec<bool> = vec![];
            unsafe {
                result.as_mut().map(|qr| {
                    qr.error = 1;
                    qr.data = data.as_mut_ptr();
                });
            }
        }
    };
    unsafe { result.as_ref().map_or(0, |qr| qr.error) }
}

#[no_mangle]
pub extern "C" fn EncodeBinary(input: *const u8, size: u32, ecl: u8, result: *mut QRCode) -> u8 {
    Clean(result);
    let binary: &[u8];
    unsafe{
        binary = std::slice::from_raw_parts(input, size as usize);
    }
    let ecl = ecl_from_int(ecl);
    wrap_qr_code(QrCode::encode_binary(binary, ecl), result);
    unsafe { result.as_ref().map_or(0, |qr| qr.error) }
}

#[no_mangle]
pub extern "C" fn Clean(result: *mut QRCode){
    unsafe {
        result.as_mut().map(|qr| {
            qr.error = 0;
            qr.width = 0;
            qr.height = 0;
            qr.needed_bits = 0;
            qr.possible_bits = 0;
            let _ = Box::from_raw(qr.data);
        });
    }
}

fn ecl_from_int(ecl: u8) -> QrCodeEcc{
    match ecl {
        1 => Some(QrCodeEcc::Low),
        2 => Some(QrCodeEcc::Medium),
        3 => Some(QrCodeEcc::Quartile),
        4 => Some(QrCodeEcc::High),
        _ => {None}
    }.unwrap_or(QrCodeEcc::High)
}

fn wrap_qr_code(qrcode: Result<QrCode, DataTooLong>, result: *mut QRCode){
    match qrcode {
        Ok(qr) => {
            let mut data: Vec<bool> = vec![];
            let size = qr.size();
            for y in 0..size {
                for x in 0..size {
                    data.push(qr.get_module(x, y));
                }
            }
            let boxed_data = Box::into_raw(Box::new(data));
            unsafe {
                result.as_mut().map(|qr| {
                    qr.width = size as u32;
                    qr.height = size as u32;
                    qr.data = boxed_data.as_mut().unwrap().as_mut_ptr();
                });
            }
        },
        Err(e) => {
            let bits = match e {
                DataTooLong::SegmentTooLong => (0, 0, 2),
                DataTooLong::DataOverCapacity(datalen, maxcapacity) => (datalen as u32, maxcapacity as u32, 3),
            };
            let mut data: Vec<bool> = vec![];
            unsafe {
                result.as_mut().map(|qr| {
                    qr.error = bits.2;
                    qr.data = data.as_mut_ptr();
                    qr.needed_bits = bits.0;
                    qr.possible_bits = bits.1;
                });
            }
        }
    }
}